@Echo OFF
Goto Start

REM  **************************************************************************
REM  *                                                                        *
REM  *  MAKEP4.BAT - Compiler Execution Script                                *
REM  *                                                                        *
REM  **************************************************************************

:Version
Echo .----------------------------------------------------------------------------.
Echo        P4 Pascal Compilation Toolkit for the Propeller - Version 1.00       
Echo `----------------------------------------------------------------------------'
Goto Return

REM  **************************************************************************
REM  *                                                                        *
REM  *  Useful links                                                          *
REM  *                                                                        *
REM  **************************************************************************

Propeller Forum thread        : http://forums.parallax.com/forums/default.aspx?f=25&m=246603

Scott A Moore's Home Page     : http://www.moorecad.com/standardpascal/p4.html
P4 Compiler Executable        : http://www.moorecad.com/standardpascal/pcom.exe
P4 Compiler Source Code       : http://www.moorecad.com/standardpascal/pcom.pas
P4 Interprter Source Code     : http://www.moorecad.com/standardpascal/pint.pas
Original Implementation Notes : http://www.moorecad.com/standardpascal/The_Pascal_P_Compiler_implementation_notes.pdf

Steven Pemberton's Home Page  : http://www.cwi.nl/~steven/pascal.html
"Pascal Implementation" book  : http://www.cwi.nl/~steven/pascal/book

Propeller Wiki Pascal Page    : http://propeller.wikispaces.com/Programming+in+Pascal

REM  **************************************************************************
REM  *                                                                        *
REM  *  Check if we have an explicit request for help                         *
REM  *                                                                        *
REM  **************************************************************************

:Start

If     "%1" == "//?" Goto Usage
If     "%1" == "--h" Goto Usage
If     "%1" == "--H" Goto Usage

If     "%1" == "/?"  Goto Usage
If     "%1" == "-h"  Goto Usage
If     "%1" == "-H"  Goto Usage

If     "%1" == "*"   Goto Special

If Not "%2" == ""    Goto Usage

REM  **************************************************************************
REM  *                                                                        *
REM  *  Check to see if it's an initial execution or a repeat of previous     *
REM  *                                                                        *
REM  **************************************************************************

If Not "%1" == "" Goto NewFile

If "%P4FILE%" == "" Goto Usage

%0 %P4FILE%
Goto Exit

:NewFile

REM  **************************************************************************
REM  *                                                                        *
REM  *  Check we're using the right OS                                        *
REM  *                                                                        *
REM  **************************************************************************

:CheckOS

If not "%P4TEST%" == ""   Goto OkayOs

If "%OS%" == "Windows_NT" Goto OkayOs
Goto WrongOs

:OkayOS

REM  **************************************************************************
REM  *                                                                        *
REM  *  Check we have the required executables needed                         *
REM  *                                                                        *
REM  **************************************************************************

If Not Exist Pcom.exe     Goto NoPcomExe

If Not Exist PrettyP4.exe Goto NoPrettyP4Exe

REM  **************************************************************************
REM  *                                                                        *
REM  *  Check the source Exists and compile it if we can. Remove any residue  *
REM  *  of previous compilations.                                             *
REM  *                                                                        *
REM  **************************************************************************

If Not Exist %1.pas Goto NoFilePas

Set P4FILE=%1

If Exist %1.err  Del %1.err
If Exist %1.p4   Del %1.p4
If Exist %1.p4i  Del %1.p4i
If Exist %1.p4a  Del %1.p4a
If Exist %1.p4b  Del %1.p4b
If Exist %1.lst  Del %1.lst
If Exist %1.bin  Del %1.bin
If Exist %1.spi  Del %1.spi
If Exist %1.spin Del %1.spin

If Exist %1.err  Echo Failed to delete '%1.err'
If Exist %1.p4   Echo Failed to delete '%1.p4'
If Exist %1.p4i  Echo Failed to delete '%1.p4i'
If Exist %1.p4a  Echo Failed to delete '%1.p4a'
If Exist %1.p4b  Echo Failed to delete '%1.p4b'
If Exist %1.lst  Echo Failed to delete '%1.lst'
If Exist %1.bin  Echo Failed to delete '%1.bin'
If Exist %1.spi  Echo Failed to delete '%1.spi'
If Exist %1.spin Echo Failed to delete '%1.spin'

If Exist %1.err  Goto Fail
If Exist %1.p4   Goto Fail
If Exist %1.p4i  Goto Fail
If Exist %1.p4a  Goto Fail
If Exist %1.p4b  Goto Fail
If Exist %1.lst  Goto Fail
If Exist %1.bin  Goto Fail
If Exist %1.spi  Goto Fail
If Exist %1.spin Goto Fail

Pcom %1.p4 < %1.pas > %1.err
If ErrorLevel 1 Goto FailPcom

REM  **************************************************************************
REM  *                                                                        *
REM  *  Pcom.exe doesn't do a good job of telling us if there were errors     *
REM  *  so we have to check for ourselves.                                    *
REM  *                                                                        *
REM  **************************************************************************

If Not Exist %1.err Goto NoFileErr
PrettyP4 %1.err
If ErrorLevel 1 Goto FailPrettyP4

If Not Exist %1.p4 Goto NoFileP4

REM  **************************************************************************
REM  *                                                                        *
REM  *  We now embark on the process of turning the '.p4' into what we need   *
REM  *  and want to produce.                                                  *
REM  *                                                                        *
REM  **************************************************************************

PrettyP4 %1.p4
If ErrorLevel 1 Goto FailPrettyP4
If Not Exist %1.p4a Goto NoFileP4a

PrettyP4 %1.p4a
If ErrorLevel 1 Goto FailPrettyP4
If Not Exist %1.p4b Goto NoFileP4b

PrettyP4 %1.p4b
If ErrorLevel 1 Goto FailPrettyP4
If Not Exist %1.lst Goto NoFileLst

REM  **************************************************************************
REM  *                                                                        *
REM  *  Generate code image
REM  *                                                                        *
REM  **************************************************************************

If Exist %1.spin      Del %1.spin
If Exist %1.spin      Echo Failed to delete %1.spin
If Exist Spin-P4.spin Del Spin-P4.spin
If Exist Spin-P4.spin Echo Failed to delete Spin-P4.spin

PrettyP4 %1.lst
If ErrorLevel 1 Goto FailPrettyP4

If Not Exist %1.bin      Goto NoFileBin
If Not Exist %1.spi      Goto NoFileSpi
If Not Exist Spin-P4.bin Goto NoSpinP4Bin
If Not Exist Spin-P4.spi Goto NoSpinP4Spi

REM  **************************************************************************
REM  *                                                                        *
REM  *  Rename the .spi file to .spin                                         *
REM  *                                                                        *
REM  **************************************************************************

Ren %1.spi %1.spin
If Not Exist %1.spin Goto NoFileSpin

Ren Spin-P4.spi Spin-P4.spin
If Not Exist Spin-P4.spin Goto NoSpinP4Spin

REM  **************************************************************************
REM  *                                                                        *
REM  *  If we got this far a cry of "Yipee!" can spread across the land       *
REM  *                                                                        *
REM  **************************************************************************

Echo.
Echo Compilation of '%1.pas' was successful - No errors detected
Echo.
Goto Exit

REM  **************************************************************************
REM  *                                                                        *
REM  *  Usage Instructions                                                    *
REM  *                                                                        *
REM  **************************************************************************

:Usage
If Not "%1" == "" Shift
If Not "%1" == "" Goto Usage
Echo.
Goto Version
:Return
If Not "%1" == "" Goto ReturnNoPcomExe
Echo.
Echo Usage : MAKEP4 {filename} {/?} {-h} {-H}
Echo.
Echo Provide the filename without the '.pas' extension but the source file must
Echo have a '.pas' extension. Once a valid filename has been provided it will be
Echo remembered and does not need to be repeated; simply type 'MAKEP4' again.
Echo.
Echo The 'Pcom.exe' P4 Compiler and the 'PrettyP4.exe' linker must both exist in
Echo the same directory as this 'MakeP4.bat' file.
Echo.
Echo When compilation successfully completes 'Spin-P4.spin' and 'Spin-P4.bin'
Echo files will be produced. These are the files used by the AiChip_P4_XXX.spin
Echo Virtual Machine. Switch to the Propeller Tool, compile AiChip_P4_XXX.spin,
Echo download and the compiled Pascal program will be executed on the Propeller.
Echo.
Echo The following files should exist in this directory ...
Echo.
Echo Compilation Tools : MakeP4.bat, Pcom.exe, PrettyP4.exe
Echo Demo Source Code  : Test.pas ( plus the compiled Test.lst and Test.p4? )
Echo Virtual Machine   : AiChip_P4_XXX.spin ( plus Spin-P4.spin and Spin-P4.bin )
Echo.
Goto Exit

REM  **************************************************************************
REM  *                                                                        *
REM  *  It's a special case command "MAKEP4 * {command}                       *
REM  *                                                                        *
REM  **************************************************************************

:Special

If "%2" == ""     Goto Usage

If "%2" == "/?"   Goto SpecUsage
If "%2" == "-h"   Goto SpecUsage
If "%2" == "-H"   Goto SpecUsage

If "%2" == "get" Goto GetPcomExe
If "%2" == "geT" Goto GetPcomExe
If "%2" == "gEt" Goto GetPcomExe
If "%2" == "gET" Goto GetPcomExe
If "%2" == "Get" Goto GetPcomExe
If "%2" == "GeT" Goto GetPcomExe
If "%2" == "GEt" Goto GetPcomExe
If "%2" == "GET" Goto GetPcomExe

If "%2" == "dir" Goto CheckDir
If "%2" == "diR" Goto CheckDir
If "%2" == "dIr" Goto CheckDir
If "%2" == "dIR" Goto CheckDir
If "%2" == "Dir" Goto CheckDir
If "%2" == "DiR" Goto CheckDir
If "%2" == "DIr" Goto CheckDir
If "%2" == "DIR" Goto CheckDir

:NoSpecial
Echo.
Echo Unknown "MAKEP4 * %2" command

:SpecUsage
Echo.
Echo MAKEP4 * DIR       Check directory configuration
Echo MAKEP4 * GET       Download 'pcom.exe'
Echo.
Goto Exit

REM  **************************************************************************
REM  *                                                                        *
REM  *  Check Directory Configuration                                         *
REM  *                                                                        *
REM  **************************************************************************

:CheckDir
Echo.
Echo Directory check ...
Echo.
Set P4FILE=

If Not Exist Pcom.exe           Set P4FILE=*
If Not Exist Pcom.exe           Echo Pcom.exe            *** MISSING ***   Compiler Executable
If     Exist Pcom.exe           Echo Pcom.exe                              Compiler Executable

If Not Exist PrettyP4.exe       Set P4FILE=*
If Not Exist PrettyP4.exe       Echo PrettyP4.exe        *** MISSING ***   Formatter, Linker and Assembler
If     Exist PrettyP4.exe       Echo PrettyP4.exe                          Formatter, Linker and Assembler

If Not Exist MakeP4.bat         Set P4FILE=*
If Not Exist MakeP4.bat         Echo MakeP4.bat          *** MISSING ***   Primary Compilation Script
If     Exist MakeP4.bat         Echo MakeP4.bat                            Primary Compilation Script

If Not Exist Make.bat           Set P4FILE=*
If Not Exist Make.bat           Echo Make.bat            *** MISSING ***   Alternative Compilation Script
If     Exist Make.bat           Echo Make.bat                              Alternative Compilation Script

If Not Exist Test.pas           Set P4FILE=*
If Not Exist Test.pas           Echo Test.pas            *** MISSING ***   Example Program Source Code
If     Exist Test.pas           Echo Test.pas                              Example Program Source Code

If Not Exist Test.lst           Set P4FILE=*
If Not Exist Test.lst           Echo Test.lst            *** MISSING ***   Example Program Listing File
If     Exist Test.lst           Echo Test.lst                              Example Program Bytecode Listing File

If Not Exist Pcom.lst           Set P4FILE=*
If Not Exist Pcom.lst           Echo Pcom.lst            *** MISSING ***   Pcom Compiler Bytecode Listing File
If     Exist Pcom.lst           Echo Pcom.lst                              Pcom Compiler Bytecode Listing File

If Not Exist AiChip_P4_???.spin Set P4FILE=*
If Not Exist AiChip_P4_???.spin Echo AiChip_P4_???.spin  *** MISSING ***   P4 Virtual Machine
If     Exist AiChip_P4_???.spin Echo AiChip_P4_???.spin                    P4 Virtual Machine

If Not Exist Spin-P4.spin       Set P4FILE=*
If Not Exist Spin-P4.spin       Echo Spin-P4.spin        *** MISSING ***   P4 Virtual Machine Opcode Definitions
If     Exist Spin-P4.spin       Echo Spin-P4.spin                          P4 Virtual Machine Opcode Definitions

If Not Exist Spin-P4.bin        Set P4FILE=*
If Not Exist Spin-P4.bin        Echo Spin-P4.bin         *** MISSING ***   Example Program P4 Opcode Image
If     Exist Spin-P4.bin        Echo Spin-P4.bin                           Example Program P4 Opcode Image

Echo.
If not "%P4FILE%" == ""         Echo FAILED : Compilation Toolkit is not complete
If     "%P4FILE%" == ""         Echo Passed : All Compilation Toolkit components present
Set P4FILE=
Echo.
Goto Exit

REM  **************************************************************************
REM  *                                                                        *
REM  *  Download Pcom.exe                                                     *
REM  *                                                                        *
REM  **************************************************************************

:GetPcomExe
Echo.
Echo Attempting to download the P4 compiler 'pcom.exe'. If this does not work you
Echo will have to manually download the compiler from ...
Echo.
Echo    Start http://www.moorecad.com/standardpascal/pcom.exe
Echo.
Start http://www.moorecad.com/standardpascal/pcom.exe
Goto Exit

REM  **************************************************************************
REM  *                                                                        *
REM  *  Error Reports                                                         *
REM  *                                                                        *
REM  **************************************************************************

:WrongOs
Echo.
Echo The 'Pcom.exe' compiler appears to be incompatible with your operating system
Echo.
If Not "%OS%" == "" Echo Your operating system seems to be : %OS%
If Not "%OS%" == "" Echo.
Echo The 'Pcom.exe' compiler requires Windows XP to execute
Echo.
Echo If you want to attempt compilation regardless of your operating system, you can
Echo circumvent future checking by using "SET P4TEST=NO".
Echo.
Goto Exit

:NoPcomExe
Echo.
Goto Version
:ReturnNoPcomExe
Echo.
Echo The compiler executable 'Pcom.exe' is not in current directory.
Echo.
Echo The compiler needs to be downloaded separately to the P4 Pascal Compilation
Echo Toolkit and can be manually downloaded from ...
Echo.
Echo     http://www.moorecad.com/standardpascal/pcom.exe
Echo.
Echo Alternatively you can attempt to download automatically using "MAKEP4 * GET"
Echo.
Echo Use "MAKEP4 * DIR" to check what is missing and what is corrctly present.
Echo.
Goto Exit

:NoPrettyP4Exe
Echo.
Echo Executable 'PrettyP4.exe' not in current directory
Echo.
Echo Use "MAKEP4 * DIR" to check what is missing and what is corrctly present.
Echo.
Goto Exit

:NoFilePas
If Exist %1 Goto GotFileXxx
Echo.
Echo File '%1.pas' does not exist
Echo.
Set P4FILE=
Goto Exit

:GotFileXxx
Echo.
Echo Cannot compile '%1'
Echo.
Echo Source file must have a '.pas' extension ( but do not include the extension ! )
Echo.
Goto Exit

:FailPcom
Echo.
Echo The 'Pcom.exe' compiler failed to compile '%1.pas'
Echo.
Goto Fail

:NoFileErr
Echo.
Echo The 'Pcom.exe' compiler failed to create '%1.err'
Echo.
Goto Fail

:NoFileP4
Echo.
Echo The 'Pcom.exe' compiler failed to create '%1.p4' - Check '%1.err'
Echo.
Goto Fail

:NoFileP4a
Echo.
Echo The 'PrettyP4.exe' linker failed to create '%1.p4a'
Echo.
Goto Fail

:NoFileP4b
Echo.
Echo The 'PrettyP4.exe' linker failed to create '%1.p4b'
Echo.
Goto Fail

:NoFileLst
Echo.
Echo The 'PrettyP4.exe' linker failed to create '%1.lst'
Echo.
Goto Fail

:NoFileBin
Echo.
Echo The 'PrettyP4.exe' linker failed to create '%1.bin'
Echo.
Goto Fail

:NoFileSpi
Echo.
Echo The 'PrettyP4.exe' linker failed to create '%1.spi'
Echo.
Goto Fail

:NoSpinP4Bin
Echo.
Echo The 'PrettyP4.exe' linker failed to create 'Spin-P4.bin'
Echo.
Goto Fail

:NoSpinP4Spi
Echo.
Echo The 'PrettyP4.exe' linker failed to create 'Spin-P4.spi'
Echo.
Goto Fail

:NoFileSpin
Echo.
Echo Failed to rename '%1.spi' to '%1.spin'
Echo.
Goto Fail

:NoSpinP4Spin
Echo.
Echo Failed to rename 'Spin-P4.spi' to 'Spin-P4.spin'
Echo.
Goto Fail

:FailPrettyP4
Echo.

:Fail
Echo Compilation of '%1.pas' failed
Echo.

:Exit

REM  **************************************************************************
REM  *                                                                        *
REM  *  End of MAKEP4.BAT batch file                                          *
REM  *                                                                        *
REM  **************************************************************************

