
REM  **************************************************************************
REM  *                                                                        *
REM  *  PRETTYP4.BAS - P4 Pascal Formatter, Linker and Assembler              *
REM  *                                                                        *
REM  **************************************************************************

$EVENT OFF

$STACK 20*1024

%TRUE=-1
%FALSE=0

dim nam$(16000)
dim lab%(16000)
dim num&(16000)
dim txt$(16000)

crlf$=CHR$(&hD)+chr$(&hA)

f$ = ltrim$(rtrim$(command$))

i& = instr(f$,";")
if i& <> 0 then
  f$=rtrim$(left$(f$,i&-1))
end if

if command$="" or instr(command$,"/?") <> 0 or left$(ucase$(command$),2)="-H" or left$(ucase$(command$),3)="--H" then
  print
  print ".----------------------------------------------------------------------------."
  print " AiChip Industries P4 Pascal Formatter, Linker and Assembler - Version 1.00 "
  print "`----------------------------------------------------------------------------'"
  print
  print "Usage : PRETTYP4 filename.ext {/?} {-h} {-H}"
  print
  print "AiChip Industries P4 Pascal Formatter, Linker and Assembler is a component"
  print "part of the P4 Pascal Compilation Toolkit for the Propeller and would not"
  print "normally be used except when invoked by the Compilation Toolkit."
  print
  if dir$("MAKEP4.BAT")<>"" Then
    print "Please use 'MAKEP4 /?' to obtain further information on using the P4 Pascal"
    print "Compilation Toolkit."
  else
    print "No 'MakeP4.bat' file exists in the current directory so the full Compilation"
    print "Toolkit may not have been completely installed."
  end if
  end 1
end if

i& = instr(f$,"/")
if i& <> 0 then
  f$=rtrim$(left$(f$,i&-1))
end if

i&=instr(f$,".")
if i& > 8 then
  Call Failed("Filename '"+f$+"' too long - Long filenames are not supported" )
else
  if instr(i&+1,f$,".") <> 0 or instr(f$," ") <> 0 then
    Call Failed("Invalid filename '"+f$+"' - Long filenames are not supported")
  else
    if instr(f$,"*") <> 0 or instr(f$,"?") <> 0 then
      Call Failed("Invalid filename '"+f$+"' - Wildcard filenames are not supported")
    end if
  end if
end if

if not abort then
  if dir$(f$)="" then
    Call Failed("File '"+f$+"' does not exist")
  end if
end if

if abort then end 1

first = %TRUE
open f$ for input as #1
while not eof(1)
  line input #1,l$
  l$=rtrim$(l$)
  if l$<>"" then
    if left$(l$,1)=";" then
      if instr(l$,".lst ")<>0 then
        gosub Generate
      else
        if instr(l$,".p4b ")<>0 then
          gosub Assemble
        else
          gosub Compact
        end if
      end if
    else
      if left$(ltrim$(l$),2)="1 " then
        gosub ExtractSource
      else
        gosub Reformat
      end if
    end if
  end if
wend
close#1
if abort then end 1
system

REM  **************************************************************************
REM  *                                                                        *
REM  *  Extract Source
REM  *                                                                        *
REM  **************************************************************************

ExtractSource:

  close#1

  heldL$=""
  count&=0
  open f$ for input as #1
  while not eof(1)
    s&=seek(1)
    line input#1,l$
    l$=rtrim$(l$)
    if instr(left$(l$,16),"*") <>0 then
      if count&=0 then
        print
        i&=instr(f$,".")
        print "Errors found in '"+left$(f$,i&-1)+".pas' ..."
        print
      end if
      count& = count&+1
      if count& <= 8 then
        if count&=8 then
          print
          print "Too many errors - See '"+f$+"' file"
        else
          print heldL$
          print l$
        end if
      end if
      abort = %TRUE
    end if
    if not abort then
      if len(l$)<>0 then
        i&=val(ltrim$(mid$(l$,7,9)))
        if i& <= 16000 then
          if num&(i&) = 0 then
            num&(i&)=s&+1
          end if
          if i& > top& then
            top&=i&
          end if
        end if
      end if
    end if
    heldL$ = l$
  wend
  close#1

  if abort then
    print
  else
    if instr(ucase$(command$),"/SOURCE")<>0 then
      i&=instr(f$,".")
      open left$(f$,i&-1)+".p4i" for output as #2
      for i&=0 to top&
        s& = num&(i&)
        if s&<>0 then
          print#2,ltrim$(str$(i&))
          open F$ for input as #1
          seek#1,s&-1
          line input#1,l$
          if len(l$)>16 then
            l$=right$(l$,len(l$)-16)
            if rtrim$(l$)<>"" then
              print#2,l$
            end if
          end if
          while not eof(1)
            line input#1,l$
            l$=rtrim$(l$)
            if len(l$)<>0 then
              if val(ltrim$(mid$(l$,7,9)))=i& then
                if len(l$)>16 then
                  l$=right$(l$,len(l$)-16)
                  if rtrim$(l$)<>"" then
                    print#2,l$
                  end if
                end if
              end if
            end if
          wend
          close#1
          print#2,""
        end if
      next
      close#2
    end if
  end if

  if abort then end 1
  system

REM  **************************************************************************
REM  *                                                                        *
REM  *  Reformat                                                              *
REM  *                                                                        *
REM  **************************************************************************

Reformat:

hadCrLf = %TRUE
needBlank = %FALSE
wasUjp = %FALSE
iCount&=3

i&=instr(f$,".")
open left$(f$,i&-1)+".p4a" for output as #2
print#2, ""
print#2, ";       File    "+left$(f$,i&-1)+".p4a created from "+f$+OnDate$
print#2, ""
print#2, ";       p4      Original p-code from pcom compilation"
print#2, ""
print#2, ";       p4a     Tidied-up formatting"
print#2, ";       p4a     Reconfigured 'lca' for compactness"
print#2, ";       p4a     Reconfigured 'ldc (...)' for clarity"
print#2, ";       p4a     Removed redundant 'xjp' operand"
print#2, ";       p4a     Removed 'i' labels"
print#2, ";       p4a     Removed 'q' labels"
print#2, ";       p4a     Added 'main' label"
print#2, ";       p4a     Added 'start ujp main' to start program"
print#2, ""
print#2, "start   ujp     main"
print#2, ""
while not eof(1)
  if first then
    first = %FALSE
  else
    line input#1,l$
  end if
  l$=rtrim$(l$)
  if l$<>"" then
    if left$(l$,1)="i" then
      if iCount& >= 0 then
        i& = val(ltrim$(right$(l$,len(l$)-1)))
        if icount& <> i& then
          Call Failed("Lost Instruction Count Synchronisation"+str$(icount&)+" "+l$)
          iCount&=i&
        end if
      end if
    else
      if l$<>firstQ$ then
        if l$="q" then
          firstQ$="q"
          l$="main"
          iCount&=-1
        end if
        if left$(l$,1)=" " then
          if iCount& >= 0 then
            iCount&=iCount&+1
          end if
          if needBlank then
            print#2, ""
            needBlank = %FALSE
          end if
          l$=ltrim$(l$)
          select case left$(l$,4)
            case "lca'"
              l$=right$(l$,len(l$)-4)
              l$=left$(l$,len(l$)-1)
              l$="lca"+str$(Len(l$))+str$(len(rtrim$(l$)))+" '"+rtrim$(l$)+"'"
            case "ldc("
              l$=left$(l$,3)+" "+right$(l$,len(l$)-3)
          end select
          if left$(l$,4)="ujp " then
            wasUjp = %true
          else
            if wasUjp then
              print#2, ""
              wasUjp = %false
              hadCrLf=%TRUE
            end if
          end if
          i& = instr(l$," ")
          if i&<>0 then
            r$ = ltrim$(right$(l$,len(l$)-i&))
            l$ = rtrim$(left$(l$,i&-1))
            select case l$
              case "lca","ldcc"
              case else
                count&=0
                o$ = ""
                w$ = ""
                for i&=1 to len(r$)
                  select case mid$(r$,i&,1)
                    case " "
                      if right$(o$,1)<>" " then
                        o$=o$+" "
                        count&=count&+1
                      end if
                      lastW$=w$
                      w$=""
                    case ")"
                      if right$(o$,1)<>" " then
                        o$=o$+" "
                      end if
                      o$=o$+")"
                      lastW$=w$
                      w$=""
                   case else
                      w$=w$+mid$(r$,i&,1)
                      o$=o$+mid$(r$,i&,1)
                  end select
                next
                r$=o$
                if l$="ldc" then
                  if lastw$="(" then
                    lastw$="0"
                  end if
                  r$=ltrim$(str$(val(lastw$)+1))+str$(count&)+" "+r$
                end if
            end select
            select case l$
              case "ujp","fjp"
                i&=instr(r$," ")
                if i& <> 0 then
                  r$ = left$(r$,i&-1)+ltrim$(right$(r$,len(r$)-i&))
                end if
              case "xjp"
                r$ = ""
              case "ent","cup"
                i&=instr(r$," ")
                if i& <> 0 then
                  i&=instr(i&+1,r$," ")
                  if i&<>0 then
                    r$ = left$(r$,i&-1)+ltrim$(right$(r$,len(r$)-i&))
                  end if
                end if
            end select
            if len(l$)<>0 Then
              l$=left$(l$+space$(8),8)+r$
            end if
          end if
          print#2,Tab(9);l$
          hadCrLf = %TRUE
          select case l$
            case "retp","reti","retb","retc"
              print#2, ""
              needBlank = %FALSE
          end select
        else
          if wasUjp then
            print#2, ""
            wasUjp = %false
            hadCrlf = %TRUE
          end if
          if not hadCrLf then
            print#2, ""
          end if
          i& = instr(l$," ")
          if i&=0 then
            if needBlank then
              print#2, ""
              needBlank = %FALSE
            end if
            print#2,l$;
            hadCrlF = %FALSE
          else
            r$ = ltrim$(right$(l$,len(l$)-i&))
            l$ = rtrim$(left$(l$,i&-1))
            l$=l$+r$
            i& = instr(l$,"=")
            if i& = 0 then
              if needBlank then
                print#2, ""
                needBlank = %FALSE
              end if
              print#2,l$;
              hadCrLf = %FALSE
            else
              r$ = ltrim$(right$(l$,len(l$)-i&))
              l$ = rtrim$(left$(l$,i&-1))
              l$ = l$+" = "+r$
              print#2,l$
              hadCrLf = %TRUE
              needBlank = %TRUE
            end if
          end if
        end if
      end if
    end if
  end if
wend
if not hadCrLf then
  print#2, ""
end if
print#2, ""
close#2
close#1
if abort then end 1
system

REM  **************************************************************************
REM  *                                                                        *
REM  *  Compact                                                               *
REM  *                                                                        *
REM  **************************************************************************

Compact:

top&=0
top&=top&+1
nam$(top&)="chka    0 2147483647"
txt$(top&)=""
num&(top&)=0
lab%(top&)=4
top&=top&+1
nam$(top&)="chka    1 2147483647"
txt$(top&)=""
num&(top&)=0
lab%(top&)=4

needBlank = %TRUE
labelCount&=0
i&=instr(f$,".")
open left$(f$,i&-1)+".p4b" for output as #2
print#2, ""
print#2, ";       File    "+left$(f$,i&-1)+".p4b created from "+f$+OnDate$

  while not eof(1)
    line input#1,l$
    if len(l$)=0 then
      needBlank = %TRUE
    else
      if left$(l$,1)=";" then
        if needBlank then
          print#2, ""
          needBlank = %FALSE
        end if
        print#2,l$
      else
        w$=""
        if left$(l$,1) <> " " then
          i&=instr(l$," = ")
          if i& <> 0 then
            i&=i&+1
            l$=rtrim$(l$)
            top&=top&+1
            nam$(top&)=rtrim$(left$(l$,i&-1))
            num&(top&)=val(ltrim$(right$(l$,len(l$)-i&)))
            lab%(top&)=1
            l$=""
          else
            i&=instr(l$," ")
            if i&=0 then
              top&=top&+1
              nam$(top&)=l$
              lab%(top&)=2
              labelCount&=labelCount&+1
              l$=""
            else
              w$=left$(l$,i&-1)
              j&=top&
              while labelCount& <> 0
                if lab%(j&)=2 then
                  txt$(j&)=w$
                  labelCount&=labelCount&-1
                end if
                j&=j&-1
              wend
              l$=ltrim$(right$(l$,len(l$)-i&))
            end if
          end if
        else
          l$=ltrim$(l$)
        end if
        if len(l$)<>0 then
          i&=instr(l$," ")
          if i& > 0 then
            select case left$(l$,i&-1)
              case "lca", "ldc", "chka","chki"
                nam$(0)=l$
                select case left$(l$,i&-1)
                  case "chka" : lab%(0) = 4
                  case "chki" : lab%(0) = 5
                  case "lca"  : lab%(0) = 6
                  case "ldc"  : lab%(0) = 7
                end select
              i&=top&
              while i& > 0
                if lab%(i&)<>lab%(0) then
                  i&=i&-1
                else
                  if nam$(i&)<>l$ then
                    i&=i&-1
                  else
                    i&=-i&
                  end if
                end if
              wend
              if i&=0 then
                top&=top&+1
                nam$(top&)=l$
                lab%(top&)=lab%(0)
              end if
            case "ujp"
              if len(w$)<>0 then
                if w$<>"start" then
                  top&=top&+1
                  nam$(top&)=w$
                  i&=instr(l$," ")
                  txt$(top&)=ltrim$(right$(l$,len(l$)-i&))
                  lab%(top&)=3
                end if
              end if
          end select
        end if
      end if
    end if
    end if
  wend
  close#1

print#2, ""
print#2, ";       p4b     Moved constant equates to top of source"
print#2, ";       p4b     Macro defined duplicate labels"
print#2, ";       p4b     Redirected jumps to 'ujp' to final destination "
print#2, ";       p4b     Removed duplicate labels from source"
print#2, ";       p4b     Extracted operand constants"
print#2, ";       p4b     Placed operand constants at top of source"
print#2, ";       p4b     Added initial offset link to operand constants"
print#2, ";       p4b     Extended operators using operand constants"
print#2, ";       p4b     Ensured 'start ujp' is to correct destination"
print#2, ""
print#2, ";       Main Program Entry Vector"
print#2, ";"
print#2, ";       The first byte of the code indicates how many bytes into the image"
print#2, ";       the constant pool is. This allows the range boundaries for addressing"
print#2, ";       to be easily located so the P4 Virtual Machine can update them as"
print#2, ";       required. Note that only the lowest 4-bits are used as the offset."
print#2, ";"
print#2, ";       If there is no bounds checking the lowest 4-bits of the offset will"
print#2, ";       be set to zero."
print#2, ";"
print#2, ";       The top 4-bits of the first byte indicates what size storage and"
print#2, ";       stack the P4 Virtual Machine should use. A value of 0 leaves the"
print#2, ";       choice to the P4 Virtual Machine itself, a value of 2 or 4 will"
print#2, ";       select a fundamental size of word or long. This will define the"
print#2, ";       maximum size of any variable as all variables are held in memory"
print#2, ";       and on the stack using the same size. This is a fundamental design"
print#2, ";       of the original Pascal P4 compiler and runtime."
print#2, ";"
print#2, ";       Immediately after that byte is the first P4 Opcode. This is a 'ujp',"
print#2, ";       an unconditional jump to the main routine of the program. Using a"
print#2, ";       'ujp' at the start of the program allows the P4 Virtual Machine to"
print#2, ";       know where to start from regardless of where the compiler may have"
print#2, ";       placed code in the executable image."
print#2, ""
print#2, "link    byte    d1"
print#2, ""
print#2, "start   ujp     "+LookupName$("main")
print#2, ""

for j&=1 to 7
  count&=0
  for i&=1 to top&
    if lab%(i&)=j& then
      if count&=0 then
  select case j&
    case 1 : print#2, ";       Equated Constants"
             print#2, ";"
             print#2, ";       A p-code interpreter expects constants to be held in the main program"
             print#2, ";       area of the image and the p-code opcodes will load constants from there"
             print#2, ";       rather than load them as immediate operands. The P4 Virtual Machine"
             print#2, ";       opcodes however support immediate operands and this list defines the"
             print#2, ";       values which were placed in the image as constants."

    case 2 : print#2, ";       Duplicate Labels"
             print#2, ";"
             print#2, ";       The pcom compiler generates labels as it requres them and this can mean"
             print#2, ";       that some locations are given multiple labels. In order to simplify the"
             print#2, ";       source code, multiple labels are removed leaving only one label in the"
             print#2, ";       source for each opcode. This is the list of labels removed followed by"
             print#2, ";       the label to which they should now refer."

    case 3 : print#2, ";       Redirected Labels"
             print#2, ";"
             print#2, ";       The pcom compiler generates labels as it requres them and this can mean"
             print#2, ";       that some labels are at 'ujp', uncoditional jump, opcodes. Rather than"
             print#2, ";       allow execution to go from one opcode to another which will then be a"
             print#2, ";       jump to another label, the original jump is redirected to what would be"
             print#2, ";       its final destination. The following is a list of all labels of 'ujp'"
             print#2, ";       opcodes and their destinations."

    case 4 : print#2, ";       Bounds Checking - Minimum Maximum"
             print#2, ";"
             print#2, ";       The first byte is split into two nibbles to indicate the form  the"
             print#2, ";       minimum and maximum values are stored in. The 3-lsbs of each nibble"
             print#2, ";       indicate how many bytes are used to store the constant, 0 through 4"
             print#2, ";       bytes. The msb of the nibble indicates whether the constant should"
             print#2, ";       be sign extended after loading. For example"
             print#2, ";"
             print#2, ";           0?                    0"
             print#2, ";           1?  nn                0 .. 255"
             print#2, ";           2?  nn nn             0 .. 65535"
             print#2, ";           3?  nn nn nn          0 .. 16777215"
             print#2, ";           4?  nn nn nn nn       0 .. 2147483647"
             print#2, ";           9?  nn               -1 .. -128"
             print#2, ";           A?  nn nn            -1 .. -32768"
             print#2, ";           B?  nn nn nn         -1 .. -8388608"
             print#2, ";           C?  nn nn nn nn      -1 .. -2147483648"
             print#2, ";"
             print#2, ";       Note that the first two 'chka' constants will always be stored with"
             print#2, ";       a size flag of $44 and held as two longs regardless of their actual"
             print#2, ";       values. These are the address range limits and holding them in this"
             print#2, ";       form allows the P4 Virtual Machine to alter the table at run time"
             print#2, ";       which would otherwise be problematical if the values needed to be"
             print#2, ";       increased and require larger storage area."

    case 5 :

    case 6 : print#2, ";       String Constants - StringLen CharsHere 'String'"
             print#2, ";"
             print#2, ";       Strings are held in the constant pool with no size indication but with"
             print#2, ";       zero termination. The compiler ensures that when strings are printed or"
             print#2, ";       assigned that they are of the correct size, however they are truncated"
             print#2, ";       for storage so data from subsequent strings ( or other data ) could be"
             print#2, ";       inadvertently printed or stored in an array of char. When printing or"
             print#2, ";       copying a string the zero termination indicates that it and subsequent"
             print#2, ";       characters should be considered spaces. It complicates the VM slightly,"
             print#2, ";       prevents $00 being stored in a string but does reduce the size of the"
             print#2, ";       string table considerably."

    case 7 : print#2, ";       Set Constants - BitsInSet BitsHere ( Bits )"
             print#2, ";"
             print#2, ";       The 'ldc' operands consist of the number of bits required to hold the"
             print#2, ";       set followed by a number which indicates how many bits are specified"
             print#2, ";       for this set, finally followed a list of bit numbers which are active"
             print#2, ";       in this set."
             print#2, ";"
             print#2, ";       The generated image code consists of a byte value which indicates how"
             print#2, ";       many bytes are required to hold the set followed by that number of"
             print#2, ";       bytes with the active bits set. Those bytes are in msb to lsb order."

  end select
  print#2, ""
      end if
      select case lab%(i&)
        case 1
          print#2, "define  "+nam$(i&);tab(16);num&(i&)
        case 2, 3
          print#2, "define  "+nam$(i&);tab(17);txt$(i&)
        case else
          print#2, "d";ltrim$(str$(i&));tab(9);nam$(i&)
      end select
      count&=count&+1
    end if
  next
  if count&<>0 then
    print#2, ""
  end if
next

read bytecodeVersion&
print#2, ";       Program Code ( Bytecode Version"+Str$(bytecodeVersion&)+" )"

needBlank = %true

open f$ for input as #1

while not eof(1)
  line input#1,l$
  l$=rtrim$(l$)
  if len(l$)=0 then
    needBlank = %true
  else
    if left$(l$,1)<>";" then
      if left$(l$,1)<>" " then
        if instr(l$," ") <> 0 then
          if instr(l$," = ") = 0 then
            if left$(l$,5)="main " then
              print#2, ""
              print#2, ";       Main Program Start"
            end if
            if left$(l$,6)<>"start " then
              Gosub Parse
            end if
          end if
        end if
      else
        Gosub Parse
      end if
    end if
  end if
wend
close#1

print#2, ""
print#2, ";       End of Program Code"
print#2, ""
close#2

  if abort then end 1
  system

Parse:
  if needBlank then
    print#2, ""
    needBlank=%FALSE
  end if
  i&=instr(l$," ")
  while mid$(l$,i&,1)=" "
    i&=i&+1
  wend
  j&=instr(i&,l$," ")
  if j& <> 0 then
    select case mid$(l$,i&,j&-i&)
      case "lca", "ldc", "chka","chki"
        w$=right$(l$,len(l$)-i&+1)
        nam$(0)=w$
        k&=top&
        while nam$(k&)<>w$
          k&=k&-1
        wend
        while mid$(l$,j&,1)=" "
          j&=j&+1
        wend
        l$=left$(l$,j&-1)+"d"+ltrim$(str$(k&))+" "+right$(l$,len(l$)-j&+1)
      case "ujp"
        ' remove label from in front of ujp
        i&=instr(l$," ")
        mid$(l$,1)=space$(i&-1)
    end select
  end if
  print#2,l$
  return

REM  **************************************************************************
REM  *                                                                        *
REM  *  Assemble                                                              *
REM  *                                                                        *
REM  **************************************************************************

Assemble:

Gosub LoadOpcodes

i&=instr(f$,".")
open left$(f$,i&-1)+".lst" for output as #2
print#2, ""
print#2, ";       File    "+left$(f$,i&-1)+".lst created from "+f$+OnDate$
for pass&=1 to 2

  needBlank= %FALSE
  pc&=0
  if pass&=2 then
    open f$ for input as #1
  end if
  first = %FALSE
  do
    line input#1,l$
    if len(l$)=0 then
      if pass&=1 then
        print#2, ""
      end if
    else
      if left$(l$,1)=";" then
        if pass&=1 then
          print#2,l$
        end if
      else
        first = %TRUE
      end if
    end if
  loop until first = %TRUE
  while not eof(1)
    if first then
      first = false
    else
      line input#1,l$
    end if
    l$=rtrim$(l$)
    if len(l$)=0 then
      if pass&=2 then
        needBlank = %TRUE
      end if
    else
      if left$(l$,1)=";" then
        if pass&=2 then
          if needBlank then
            print#2, ""
            needBlank = %FALSE
          end if
          print#2,l$
        end if
      else
        o$=""
        k$=""
        if left$(l$,1)<>" " then
          i&=instr(l$," ")
          if left$(l$,i&-1)="define" then
            ' handle definition
            globalL$=l$
            gosub word
            gosub word
            if pass&=1 then
              top&=top&+1
              nam$(top&)=w$
              gosub word
              if instr("-+0123456789",left$(w$,1))=0 then
                txt$(top&)=w$
                lab%(top&)=0
              else
                num&(top&)=val(w$)
                lab%(top&)=1
              end if
            else
              k$=space$(8)+FNhex$(Lookup&(w$),4)
            end if
          else
            ' add label definition
            if pass&=1 then
              top&=top&+1
              nam$(top&)=left$(l$,i&-1)
              num&(top&)=pc&
              lab%(top&)=-1
            else
              wasPc& = lookup&(left$(l$,i&-1))
              if wasPc& <> pc& then
                Call Failed("Lost label synchronisation "+left$(l$,i&-1))
                print#2,""
                print#2,"FAILED "+left$(l$,i&-1)+" was "+FNhex$(wasPc&,4)+" now "+FNhex$(pc&,4)
                print#2,""
                needBlank=%FALSE
                pc& = wasPc&
              end if
            end if
            if left$(l$,1)="d" then
              Call CompileLine(right$(l$,len(l$)-i&),%TRUE)
            else
              Call CompileLine(right$(l$,len(l$)-i&),%FALSE)
            end if
          end if
        else
          Call CompileLine(l$,%FALSE)
        end if
        if pass&=1 then
          pc&=pc&+(Len(o$)\3)
        else
          %BYTES = 10
          %WIDE  = %BYTES*3
          if needBlank then
            print#2, ""
            needBlank = %FALSE
          end if
          if len(o$)<>0 then
            if len(o$) > %WIDE then
              needBlank = %TRUE
              print#2,FNhex$(pc&,4);tab(9);k$;tab(17);left$(o$,%WIDE);tab(49);l$
              o$=right$(o$,len(o$)-%WIDE)
              pc&=pc&+(%WIDE\3)
              while len(o$) > %WIDE
                print#2,FNhex$(pc&,4);tab(17);left$(o$,%WIDE)
                o$=right$(o$,len(o$)-%WIDE)
                pc&=pc&+(%WIDE\3)
              wend
              print#2,FNhex$(pc&,4);tab(17);o$
              pc&=pc&+(Len(o$)\3)
            else
              print#2,FNhex$(pc&,4);tab(9);k$;tab(17);o$;tab(49);l$
              pc&=pc&+(Len(o$)\3)
            end if
          else
            print#2,k$;tab(49);l$
          end if
        end if
      end if
    end if
  wend
  close#1
next
if needBlank then
  print#2, ""
end if
' ///
' for i&=1 to top&
'   print#2,lab%(i&),nam$(i&),hex$(num&(i&)),num&(i&),txt$(i&)
' next
' print#2, ""
close#2

if abort then end 1
system

Sub CompileLine(ll$,const)
shared pass&
shared o$
shared globalL$
shared w$
local I&
local b$
local vMin&,vMax&
local vMinSize&,vMaxSize&

  globalL$=ltrim$(ll$)
  gosub word
  select case w$
    case "chka"
      if const then
        call genbyte(&h44)
        gosub Word
        call GenLong(val(w$))
        gosub word
        call GenLong(val(w$))
      else
        call CompileOpcode
      end if
    case "chki"
      if const then
        gosub Word
        vMin& = val(w$)
        vMinSize&=NumSize&(vMin&)
        gosub word
        vMax& = val(w$)
        vMaxSize&=NumSize&(vMax&)
        call genbyte(vMinSize&*16+vMaxSize&)
        select case vMinSize& and 7
          case 1 : call genbyte(vMin&)
          case 2 : call genword(vMin&)
          case 3 : call gentrip(vMin&)
          case 4 : call genlong(vMin&)
        end select
        select case vMaxSize& and 7
          case 1 : call genbyte(vMax&)
          case 2 : call genword(vMax&)
          case 3 : call gentrip(vMax&)
          case 4 : call genlong(vMax&)
        end select
      else
        call CompileOpcode
      end if
    case "ldc"
      if const then
        gosub Word
        i&=val(w$)
        if ( i& mod 8) <> 0 then
          i&=(i&\8)+1
        else
          i&=i&\8
        end if
        call genbyte(i&)
        b$=string$(i&,chr$(0))
        gosub word
        gosub word
        gosub word
        while w$<>")"
          i&=val(w$)
          mid$(b$,(i&\8)+1,1) = chr$(asc(mid$(b$,(i&\8)+1,1)) or (2^(i& mod 8)) )
          gosub word
        wend
        for i&=len(b$) to 1 step -1
          call genbyte( Asc(mid$(b$,i&,1)) )
        next
      else
        call CompileOpcode
      end if
    case "lca"
      if const then
        gosub Word
        gosub Word
        gosub Word
        for i&=2 to len(w$)-1
          call genbyte(asc(mid$(w$,i&,1)))
        next
        call genByte(0)
      else
        call CompileOpcode
      end if
    case "csp"
      gosub word
      w$="csp"+w$
      call CompileOpcode
    case "eof"
      w$="cspeof"
      call CompileOpcode
    case "ent"
      gosub word
      w$="ent"+w$
      call CompileOpcode
    case else
      call CompileOpcode
  end select
End Sub

Function NumSize&( n& )
local s&
  if n&=0 then
    s& = 0
  else
    if n& > 0 then
      if n& < 256 then
        s& = 1
      else
        if n& < 65536 then
          s& = 2
        else
          if n& < 16777216 then
            s& = 3
          else
            s& = 4
          end if
        end if
      end if
    else
      if n& >= -128 then
        s& = 1 + 8
      else
        if n& >= -32768 then
          s& = 2 + 8
        else
          s& = 4
        end if
      end if
    end if
  end if
  NumSize& = s&
End Function

Sub CompileOpcode
shared top&,nam$(),num&(),txt$(),lab%()
Shared w$,pass&,opcTop&
shared l$,o$,k$,abort,crlf$
local opcodePtr&
local opcodeMrk&
local i&
local t$
local v&
local vSize&
local opcode&
  opcodePtr&=opcTop&
  nam$(0)=w$
  while nam$(opcodePtr&) <> w$
    opcodePtr&=opcodePtr&-1
  wend
  if opcodePtr&=0 then
    Call Failed("Missing Opcode "+w$)
  end if
  opcode&=num&(opcodePtr&)
  opcodeMrk&=opcode&
  t$=txt$(opcodePtr&)
  o$=""

  for i&=1 to len(t$)
    gosub word
    v& = WordVal&(t$,i&)
    vSize&=NumSize&(v&)
    if vSize&=0 then
      vSize&=1
    end if
    select case mid$(t$,i&,1)
      case "N"
        gosub word
        while w$ <> ""
          v& = ( v& and &hF ) * 16 + ( WordVal&(t$,i&) and &hF)
          gosub word
        wend
        vSize&=1
      case "B" :
        select case vSize&
          case 1    :
          case else :
            Call Failed(ltrim$(l$)+CRLF$+"Size too big "+mid$(t$,i&,1)+" "+hex$(v&)+" /"+str$(v&))
        end select
        vSize&=1
      case "W","A","K" :
        select case vSize&
          case 1    :
          case 2    :
          case else :
            Call Failed(ltrim$(l$)+CRLF$+"Size too big "+mid$(t$,i&,1)+" "+hex$(v&)+" /"+str$(v&))
        end select
        vSize&=2
      case "L" :
        vSize&=4
      case "t","f"
        select case v&
          case 0 : opcode& = opcode&-1
          case 1 :
          case else :
            Call Failed(ltrim$(l$)+CRLF$+"Size too big "+mid$(t$,i&,1)+" "+hex$(v&)+" /"+str$(v&))
        end select
        vSize&=0
      case "b"
        select case vSize&
          case 1    :
            select case v&
              case 0 : opcode& = opcode&-2 : vSize& = 0
              case 1 : opcode& = opcode&-1 : vSize& = 0
            end select
          case else :
            Call Failed(ltrim$(l$)+CRLF$+"Size too big "+mid$(t$,i&,1)+" "+hex$(v&)+" /"+str$(v&))
        end select
      case "w"
        select case vSize&
          case 1    : opcode& = opcode&-1
          case 2    :
          case else :
            Call Failed(ltrim$(l$)+CRLF$+"Size too big "+mid$(t$,i&,1)+" "+hex$(v&)+" /"+str$(v&))
        end select
      case "l"
        select case vSize&
          case 1    : opcode& = opcode&-2
          case 2    : opcode& = opcode&-1
        end select
      case else
        Call Failed(ltrim$(l$)+crlf$+"Unknown operand type "+mid$(t$,i&,1))
    end select
    select case vSize&
      case 0    :
      case 1    : call genbyte(v&)
      case 2    : call genword(v&)
      case else : call genlong(v&)
    end select
  next

  if opcode& >=0 then
    o$=FNhex$(opcode&,2)+" "+o$
  end if

  i&=opcodePtr&+opcode&-opcodeMrk&
  t$=nam$(i&)
  if lab%(i&) = -3 then
    t$=left$(t$,len(t$)-1)+"x"
  end if
  k$=txt$(i&)
  if len(k$)=0 then
    k$=t$
  else
    if k$=ucase$(k$) then
      if instr("01",left$(k$,1))=0 then
        k$=t$
      else
        k$=left$(t$+space$(5),5)+k$
      end if
    else
      k$=left$(t$+space$(5),5)+k$
    end if
  end if
  k$=lcase$(k$)
end sub

Function WordVal&(t$,i&)
shared w$,pass&
local v&
    if w$="" then
      v&=0
    else
      select case instr("'-+0123456789",left$(w$,1))
        case 0
          if pass&=1 then
            if mid$(t$,i&,1)=lcase$(mid$(t$,i&,1)) then
              v& = Lookup&(w$)
            else
              v&=0
            end if
          else
            v& = Lookup&(w$)
          end if
        case 1
          v& = asc(mid$(w$,2,1))
        case else
          v&= val(w$)
      end select
    end if
  WordVal&=v&
end function

Sub GenByte(n&)
shared o$
  o$ = o$+FNhex$(n&,2)+" "
end sub

Sub GenWord(n&)
shared o$
local h$
  h$=Fnhex$(n&,4)
  o$ = o$+Left$(h$,2)+" "+right$(h$,2)+" "
end sub

Sub GenTrip(n&)
shared o$
local h$
  h$ = MKL$(n&)
  o$ = o$+FNhex$(asc(mid$(h$,3,1)),2)+" "
  o$ = o$+FNhex$(asc(mid$(h$,2,1)),2)+" "
  o$ = o$+FNhex$(asc(mid$(h$,1,1)),2)+" "
end sub

Sub GenLong(n&)
shared o$
local h$
  h$ = MKL$(n&)
  o$ = o$+FNhex$(asc(mid$(h$,4,1)),2)+" "
  o$ = o$+FNhex$(asc(mid$(h$,3,1)),2)+" "
  o$ = o$+FNhex$(asc(mid$(h$,2,1)),2)+" "
  o$ = o$+FNhex$(asc(mid$(h$,1,1)),2)+" "
end sub

Def FNhex$(n&,w&)
  FNhex$=right$("0000"+hex$(n&),w&)
End Def

Word:
  if left$(globalL$,1)="'" then
    w$=globalL$
    globalL$=""
  else
    i&=Instr(globalL$," ")
    if i&=0 then
      w$=globalL$
      globalL$=""
    else
      w$=left$(globalL$,i&-1)
      globalL$=ltrim$(right$(globalL$,len(globalL$)-i&))
    end if
  end if
  return

REM  **************************************************************************
REM  *                                                                        *
REM  *  Generate Bytecode                                                     *
REM  *                                                                        *
REM  **************************************************************************

Generate:

Gosub LoadOpcodes

for pass&=1 to 2

  if pass&=1 then
    i&=instr(f$,".")
    open left$(f$,i&-1)+".bin" for output as #2
  else
    open f$ for input as #1
    open "Spin-P4.bin" for output as #2
  end if
  while not eof(1)
    line input#1,l$
    l$=rtrim$(l$)
    if len(l$)<>0 then
      if left$(l$,1)<>";" then
        if left$(l$,1)<>" " then
          o$=""
          l$=right$(l$,len(l$)-17+1)
          i&=1
          while i& > 0
            if i& > len(l$) then
              i&=0
            else
              h$=mid$(l$,i&,2)
              if h$="  " then
                i&=0
              else
                o$=o$+chr$(val("&h"+h$))
                i&=i&+3
              end if
            end if
          wend
          print#2, o$;
        end if
      end if
    end if
  wend
  close#1
  close#2
  i&=instr(f$,".")
  if pass&=1 then
    open left$(f$,i&-1)+".spi" for output as #2
  else
    open "Spin-P4.spi" for output as #2
  end if
  print#2, ""
  print#2, "PUB SourceFilename"
  print#2, ""
  print#2, "  return String( "+chr$(&h22)+left$(f$,i&-1)+chr$(&h22)+" )"
  print#2, ""
  print#2, "PUB BytecodeStart"
  print#2, ""
  print#2, "  return @P4_BytecodeStart"
  print#2, ""
  print#2, "DAT"
  print#2, ""
  if pass&=1 then
    print#2, "  P4_BytecodeStart FILE "+chr$(&h22)+left$(f$,i&-1)+".bin"+chr$(&h22)
  else
    print#2, "  P4_BytecodeStart FILE "+chr$(&h22)+"Spin-P4.bin"+chr$(&h22)
  end if
  print#2, ""
  print#2, "CON"
  print#2, ""
  print#2, "  BYTECODE_VERSION";tab(25);"="+str$(bytecodeVersion&)
  print#2, ""
  print#2, "  FILE_INPUT";TAB(25);"= 5"
  print#2, "  FILE_OUTPUT";TAB(25);"= 6"
  print#2, "  FILE_PRD";TAB(25);"= 7"
  print#2, "  FILE_PRR";TAB(25);"= 8"
  print#2, ""
  for i&=1 to top&
    if lab%(i&)=-2 then
      if num&(i&) >= 0 then
        opc$=ucase$(nam$(i&))
        opr$=ucase$(txt$(i&))
        if opr$="" then
          w$=""
        else
          w$="_"+opr$
          opr$="  ' "+opr$
        end if
        print#2, "  Opc_"+opc$+w$;Tab(25);"= $"+FNhex$(num&(i&),2)+opr$
      end if
    end if
  next
  count&=0
  indent$="  "
  print#2, ""
  print#2, "PUB OpcodeNamePtr( p4opcode )"
  for i&=1 to top&
    if lab%(i&)=-2 then
      if num&(i&) >= 0 then
        if (count& mod 64 ) = 0 then
          if count&<>0 then
            print#2,indent$+"other:"
            indent$=indent$+"  "
          end if
          print#2,indent$+"case p4opcode"
          indent$=indent$+"  "
        end if
        opc$=ucase$(nam$(i&))
        opr$=ucase$(txt$(i&))
        l$=opc$
        if opr$="" then
          w$=""
        else
          w$="_"+opr$
          l$=left$(left$(l$+space$(5),5)+opr$+space$(8),8)
        end if
        print#2,indent$+"Opc_"+opc$+w$;tab(25);": result := String("+chr$(&h22)+l$+Chr$(&h22)+")";TAB(57);"' $"+FNhex$(num&(i&),2)
        count&=count&+1
      end if
    end if
  next
  print#2,indent$+"other";Tab(25);": result := String("+chr$(&h22)+"Unknown"+Chr$(&h22)+")"
  print#2, ""
  close#2
next

if instr(ucase$(command$),"/TEMPLATE") <> 0 then

  open "template.txt" for output as #2
  print#2,"PRI P4Handler"
  print#2,""
  print#2,"  if opcode =< p4#Opc_"+ucase$(nam$(cspTop&))
  print#2,"    CspHandler"
  print#2,"  else"
  print#2,"    if opcode =< p4#Opc_"
  print#2,"      case opcode"
  count&=0
  for i&=cspTop&+1 to top&
    if lab%(i&)=-2 and num&(i&) >=0 then
      if count&<>0 and (count& mod 64)=0 then
        print#2,"    else"
        print#2,"      case opcode"
      end if
      l$=txt$(i&)
      if l$<>"" then
        l$="_"+l$
      end if
      print#2,"        p4#Opc_"+ucase$(nam$(i&)+l$)+":"
      count&=count&+1
    end if
  next
  print#2,""
  print#2,"PRI CspHandler"
  print#2,""
  print#2,"  case opcode"
  for i&=1 to cspTop&
    if lab%(i&)=-2 then
      print#2,"    p4#Opc_"+ucase$(nam$(i&))+":    NotImplemented"
    end if
  next
  print#2,""
  close#2

end if

if abort then end 1
system

REM  **************************************************************************
REM  *                                                                        *
REM  *  Utility Code                                                          *
REM  *                                                                        *
REM  **************************************************************************

Function Lookup&(n$)
shared top&,nam$(),num&(),txt$(),lab%(),pass&,abort
Local i&
  nam$(0)=n$
  lab%(0)=1
  num&(0)=0
  txt$(0)=""
  do
    i&=top&
    while nam$(i&)<>nam$(0)
      i&=i&-1
    wend
    if i&=0 then
      Call Failed("Crashed "+n$+" -> "+nam$(0))
    end if
    if lab%(i&) = 0 then
      nam$(0) = txt$(i&)
    end if
 loop until lab%(i&)<>0
 lookup&=num&(i&)
end function

Function LookupName$(n$)
shared top&,nam$(),num&(),txt$(),lab%(),pass&
Local i&
    nam$(0)=n$
    lab%(0)=1
    num&(0)=0
    txt$(0)=""
    do
      i&=top&
      while nam$(i&)<>nam$(0)
        i&=i&-1
      wend
      if i&=0 then
        lab%(0)=1
      else
        if lab%(i&) <> 1 then
          nam$(0) = txt$(i&)
        end if
      end if
   loop until lab%(i&) = 1
   lookupName$=nam$(0)
end function


Function OnDate$
local d$
local t$
  d$ = date$
  t$ = time$
  while d$+t$<>date$+time$
    d$ = date$
    t$ = time$
  wend
  OnDate$=" ( "+right$(d$,4)+"-"+mid$(d$,4,2)+"-"+Left$(d$,2)+" "+t$+" )"
End Function

REM  **************************************************************************
REM  *                                                                        *
REM  *  Bytecode / P-Code Definitions                                         *
REM  *                                                                        *
REM  **************************************************************************

LoadOpcodes:

  top&=0
  opcTop&=-3
  opcHld&=-1
  csptop&= 1
  read bytecodeVersion&
  read l$
  while len(l$) <> 0
    i& = instr(l$," ")
    if i&=0 then
      opc$=l$
      opr$=""
    else
      opc$=left$(l$,i&-1)
      opr$=ltrim$(right$(l$,len(l$)-i&))
    end if
    if right$(opc$,3)="*+r" then
      opc$=left$(opc$,len(opc$)-3)
      Call OpcodeHld(opc$+"a",opr$)
      Call OpcodeHld(opc$+"b",opr$)
      Call OpcodeHld(opc$+"c",opr$)
      Call OpcodeHld(opc$+"i",opr$)
      Call OpcodeHld(opc$+"s",opr$)
      Call OpcodeHld(opc$+"r",opr$)
      Call OpcodeLhs(opc$+"x",opr$) ' <-- Needed, must be Lhs
    else
      if right$(opc$,1)="*" then
        opc$=left$(opc$,len(opc$)-1)
        Call OpcodeHld(opc$+"a",opr$)
        Call OpcodeHld(opc$+"b",opr$)
        Call OpcodeHld(opc$+"c",opr$)
        Call OpcodeHld(opc$+"i",opr$)
        Call OpcodeHld(opc$+"s",opr$)
        Call OpcodeLhs(opc$+"x",opr$) ' <-- Needed, must be Lhs
      else
        Call OpcodeLhs(opc$,opr$)
      end if
    end if
    read l$
  wend
  opcTop& = top&
  return

Sub OpcodeHld(opc$,opr$)
Shared opcTop&
Shared opcHld&
  opcHld& = opcTop&
  Call OpcodeLhs(opc$,opr$)
  opcTop& = opcHld&
  opcHld& = -1
End Sub

Sub OpcodeLhs(opc$,opr$)
   select case left$(opr$,1)
     case "b"  : Call OpcodeRhs(opc$,"z"+right$(opr$,len(opr$)-1))
                 Call OpcodeRhs(opc$,"o"+right$(opr$,len(opr$)-1))
                 Call OpcodeRhs(opc$,"b"+right$(opr$,len(opr$)-1))
     case "w"  : Call OpcodeRhs(opc$,"b"+right$(opr$,len(opr$)-1))
                 Call OpcodeRhs(opc$,"w"+right$(opr$,len(opr$)-1))
     case "k"  : Call OpcodeRhs(opc$,"b"+right$(opr$,len(opr$)-1))
                 Call OpcodeRhs(opc$,"w"+right$(opr$,len(opr$)-1))
     case "l"  : Call OpcodeRhs(opc$,"b"+right$(opr$,len(opr$)-1))
                 Call OpcodeRhs(opc$,"w"+right$(opr$,len(opr$)-1))
                 Call OpcodeRhs(opc$,"l"+right$(opr$,len(opr$)-1))
     case "t"  : Call OpcodeRhs(opc$,"f"+right$(opr$,len(opr$)-1))
                 Call OpcodeRhs(opc$,"t"+right$(opr$,len(opr$)-1))
     case else : Call OpcodeRhs(opc$,opr$)
   end select
End Sub

Sub OpcodeRhs(opc$,opr$)
shared opchld$
   if len(opr$) <= 1 then
     Call OpcodeFinal(opc$,opr$)
   else
     select case right$(opr$,1)
       case "b"  : Call OpcodeFinal(opc$,left$(opr$,1)+"z")
                   Call OpcodeFinal(opc$,left$(opr$,1)+"o")
                   Call OpcodeFinal(opc$,left$(opr$,1)+"b")
       case "w"  : Call OpcodeFinal(opc$,left$(opr$,1)+"b")
                   Call OpcodeFinal(opc$,left$(opr$,1)+"w")
       case "l"  : Call OpcodeFinal(opc$,left$(opr$,1)+"b")
                   Call OpcodeFinal(opc$,left$(opr$,1)+"w")
                   Call OpcodeFinal(opc$,left$(opr$,1)+"l")
       case "t"  : Call OpcodeFinal(opc$,left$(opr$,1)+"f")
                   Call OpcodeFinal(opc$,left$(opr$,1)+"t")
       case else :
                   if instr(opc$,"*")<>0 then
                     opcHld& = 1
                   end if
                   Call OpcodeFinal(opc$,opr$)
                   opcHld& = -1
     end select
  end if
End Sub

Sub OpcodeFinal(opc$,opr$)
shared top&,nam$(),num&(),txt$(),lab%(),opctop&
shared opchld&,csptop&
  top&=top&+1
  nam$(top&)=opc$
  txt$(top&)=opr$
  num&(top&)=opctop&
  if opchld& >=0 then
    lab%(top&)=-3
  else
    lab%(top&)=-2
  end if
  if left$(opc$,3)="csp" then
    csptop&=top&
  end if
  opcTop&=opcTop&+1
End Sub

REM  **************************************************************************
REM  *                                                                        *
REM  *  Utility Code                                                          *
REM  *                                                                        *
REM  **************************************************************************

Sub Failed(l$)
Shared abort
  print l$
  abort = %TRUE
End Sub

REM  **************************************************************************
REM  *                                                                        *
REM  *  Bytecode - DON'T FORGET TO INCREMENT THE VERSION NUMBER !             *
REM  *                                                                        *
REM  **************************************************************************

Data 0          ' VERSION NUMBER

' These threee must be first. If adding anything, the source code above will
' have to be changed to take that into account

Data "long L"
Data "word W"
Data "byte N"

' The csp definitions should come before all others

Data "cspeln"
Data "cspeof"
Data "cspget"
Data "cspnew"
Data "cspput"
Data "csprdc"
Data "csprdi"
Data "csprln"
Data "csprst"
Data "cspsav"
Data "cspwln"
Data "cspwrb"           ' Write boolean
Data "cspwrc"
Data "cspwri"
Data "cspwrs"           ' Write string
Data "cspwru"           ' Write set ( union )

' P-Code opcodes - only one lowercase parameter per opcode allowed !

Data "abi"
Data "abr"
Data "adi"
Data "adr"
Data "and"
Data "chka   k"
Data "chkb"
Data "chkc   bB"
Data "chki   k"
Data "chr"
Data "cup    bW"
Data "deci   l"
Data "dif"
Data "dvi"
Data "dvr"
Data "ent1   w"
Data "ent2   w"
Data "equ*"
Data "equr"
Data "equm   W"
Data "fjp    A"
Data "flo"
Data "flt"
Data "geq*"
Data "geqr"
Data "geqm   W"
Data "grt*"
Data "grtr"
Data "grtm   W"
Data "inc*   l"
Data "ind*+r w"
Data "inn"
Data "int"
Data "ior"
Data "ixa    w"
Data "lao    w"
Data "lca    k"
Data "lda    Bw"
Data "ldc    k"
Data "ldcb   t"
Data "ldcc   b"
Data "ldci   l"
Data "ldcn"
Data "ldo*+r w"
Data "leq*"
Data "leqr"
Data "leqm   W"
Data "les*"
Data "lesr"
Data "lesm   W"
Data "lod*+r Bw"
Data "loda   Bw" ' This must foloow lod*+r as they are not teh same !
Data "mod"
Data "mov    w"
Data "mpi"
Data "mpr"
Data "mst    b"
Data "neq*"
Data "neqr"
Data "neqm   W"
Data "ngi"
Data "ngr"
Data "not"
Data "odd"
Data "ord*"
Data "ret*+r"
Data "retp"
Data "sbi"
Data "sbr"
Data "sgs"
Data "sqi"
Data "sqr"
Data "sro*+r w"
Data "sto*+r"
Data "stp"
Data "str*+r Bw"
Data "trc"
Data "ujc"
Data "ujp    A"
Data "uni"
Data "xjp"

Data ""

REM  **************************************************************************
REM  *                                                                        *
REM  *  End of PRETTYP4.BAS Program                                           *
REM  *                                                                        *
REM  **************************************************************************

